
#include "lib.h"



#define INDENT_MAKER
struct symtable * symtable_new()
{
    struct symtable * t = malloc(sizeof(struct symtable));
    if(t == NULL){
        fprintf(stderr,"Bug memory allocation failed\n");
        exit(1);
    }
    t->capacity = 1024;
    t->symbols = malloc(t->capacity*sizeof(struct symbol));
    if(t->symbols  == NULL){
        fprintf(stderr,"Bug memory allocation failed\n");
        exit(1);
    }
    t->temporary = 0;
    t->condition = 0;
    t->size = 0;
    init_indent();
    return t;
}

static void symtable_grow(struct symtable * t) 
{
    t->capacity += 1024;
    t->symbols = realloc(t->symbols,t->capacity*sizeof(struct symbol));
    if(t->symbols == NULL) 
    {
      fprintf(stderr,"Error attempting to grow symbol table (actual size is %d)\n",t->size);
        exit(1);
    }
}

struct symbol* symtable_const_int(struct symtable * t, long int v) 
{
    unsigned int i;
    for ( i=0 ; i<t->size; i++ ){
        if(t->symbols[i].kind == CONSTANT_INT){
            if(t->symbols[i].u.value_int == v){
                break;
            }
        }
    }
    if(i==t->size)
    {
        if(t->size==t->capacity)
          symtable_grow(t);
        struct symbol *s = &(t->symbols[t->size]);
        s->kind = CONSTANT_INT;
        s->u.value_int = v;
        ++ (t->size);
        return s;
    }
    else 
    {
        return &(t->symbols[i]);
    }
}

struct symbol* symtable_const_double(struct symtable * t, float v) 
{
    unsigned int i;
    for ( i=0 ; i<t->size; i++ ){
        if(t->symbols[i].kind == CONSTANT_FLOAT){
            if(t->symbols[i].u.value_float == v){
                break;
            }
        }
    }
    if(i==t->size)
    {
        if(t->size==t->capacity)
          symtable_grow(t);
        struct symbol *s = &(t->symbols[t->size]);
        s->kind = CONSTANT_FLOAT;
        s->u.value_float = v;
        ++ (t->size);
        return s;
    }
    else 
    {
        return &(t->symbols[i]);
    }
}

struct symbol* symtable_get(struct symtable * t, const char * id, struct symbol* func_id) 
{
    unsigned int i;
    for ( i=0 ; i<t->size; i++ )
    {
        if(strcmp(t->symbols[i].u.name,id) == 0)
        {
            if(t->symbols[i].kind == NAME_LOC || t->symbols[i].kind == ARRAY)
            { // these names are only unique whithin a function_scope
                // if the scope function is the same it's the right symbol
                if(strcmp(func_id->u.name,t->symbols[i].scope_function) == 0)
                    return &(t->symbols[i]);
            } 
            else if ((t->symbols[i].kind == NAME || t->symbols[i].kind == NAME_FLOAT || t->symbols[i].kind == LABEL) && func_id == NULL)  
            { // these names are supposed to be unique
                return &(t->symbols[i]);
            }
        }
    }
      
    return NULL;
}


struct symbol* symtable_put_name(struct symtable * t, const char * id) 
{
    if(t->size==t->capacity)
      symtable_grow(t);
    struct symbol *s = &(t->symbols[t->size]);
    s->kind = NAME;
    strcpy(s->u.name,id);
    ++ (t->size);
    return s;
}
struct symbol* symtable_put_name_float(struct symtable * t, const char * id)
{
    if(t->size==t->capacity)
      symtable_grow(t);
    struct symbol *s = &(t->symbols[t->size]);
    s->kind = NAME_FLOAT;
    strcpy(s->u.name,id);
    ++ (t->size);
    return s;
}

struct symbol* symtable_put_chaine(struct symtable * t, const char * chaine) 
{
    if(t->size==t->capacity)
      symtable_grow(t);
    struct symbol *s = &(t->symbols[t->size]);
    s->kind = CHAIN;
    strcpy(s->u.chaine,chaine);
    ++ (t->size);
    return s;
}

struct symbol* symtable_put_loc(struct symtable * t, const char * var_name, struct symbol* func_id) 
{
    if(t->size==t->capacity)
      symtable_grow(t);
    struct symbol *s = &(t->symbols[t->size]);
    s->kind = NAME_LOC;
    strcpy(s->u.name,var_name);
    strcpy(s->scope_function,func_id->u.name);
    ++ (t->size);
    return s;
}

struct symbol* symtable_put_array(struct symtable * t, const char* var_name, unsigned int dim, struct symbol** arrayDim, struct symbol* dim1size, struct symbol* func_id){
    if(t->size==t->capacity)
      symtable_grow(t);
    struct symbol *s = &(t->symbols[t->size]);

    s->kind = ARRAY;
    strcpy(s->u.arr.name,var_name);
    s->u.arr.dim = dim;
    s->u.arr.dim1size = dim1size;

    // Allocate a new array inside the structure because the argument arrayDim will be freed
    s->u.arr.size = (struct symbol**) malloc(dim*sizeof(struct symbol*));
    if(s->u.arr.size == NULL){
        fprintf(stderr,"Bug memory allocation failed\n");
        exit(1);
    }

    // Copy argument arrayDim in the array inside the structure
    for(unsigned int i=0;i<dim;i++){
        s->u.arr.size[i] = arrayDim[i];
    }

    strcpy(s->scope_function,func_id->u.name);

    ++ (t->size);
    return s;
}

struct symbol *newtemp(struct symtable * t, enum var_type ty)
{
    if(t->size==t->capacity)
      symtable_grow(t);
    struct symbol* s;
    name_t name;
    sprintf(name,"t%d",t->temporary);
    if(ty == INTEGER)
        s = symtable_put_name(t,name);
    else if( ty == FLOAT)
        s = symtable_put_name_float(t,name);
    else
        fprintf(stderr,"wrong var_type given, or do not exist yet\n");
    ++ (t->temporary);   
    return s;
}

struct symbol* symtable_put_label(struct symtable * t, const char* var_name, int cond_id, struct symbol* func_id)
{
    if(t->size==t->capacity)
      symtable_grow(t);
    struct symbol *s = &(t->symbols[t->size]);
    s->kind = LABEL;
    sprintf(s->u.name,"%s_%d",var_name,cond_id);
    strcpy(s->scope_function,func_id->u.name);
    ++ (t->size);
    return s;
}

void symtable_dump(struct symtable * t, FILE* fout)
{
    unsigned int i;
    for ( i=0 ; i<t->size; i++ )
    {
        if(t->symbols[i].kind==CONSTANT_INT)
            fprintf(fout,"#       %p = %ld\n",&(t->symbols[i]),t->symbols[i].u.value_int);
        else if(t->symbols[i].kind==CONSTANT_FLOAT)
            fprintf(fout,"#       %p = %0.2f\n",&(t->symbols[i]),t->symbols[i].u.value_float);
        else if(t->symbols[i].kind==NAME)
            fprintf(fout,"#       %p = %s (INTEGER)\n",&(t->symbols[i]),t->symbols[i].u.name);
        else if(t->symbols[i].kind==NAME_FLOAT)
            fprintf(fout,"#       %p = %s (FLOAT)\n",&(t->symbols[i]),t->symbols[i].u.name);
        else if(t->symbols[i].kind==NAME_LOC)
            fprintf(fout,"#       %p = %s (in the scope of the '%s' function type : INTEGER)\n",&(t->symbols[i]),t->symbols[i].u.name,t->symbols[i].scope_function);
        else if(t->symbols[i].kind==CHAIN)
            fprintf(fout,"#       %p = %s\n",&(t->symbols[i]),t->symbols[i].u.chaine);
        else if(t->symbols[i].kind==LABEL)
            fprintf(fout,"#       %p = %s (LABEL)\n",&(t->symbols[i]),t->symbols[i].u.name);
        else if(t->symbols[i].kind==ARRAY){
            fprintf(fout,"#       %p = (%s,",&(t->symbols[i]),t->symbols[i].u.arr.name);
        if(t->symbols[i].u.arr.size[0]->kind == NAME_LOC || t->symbols[i].u.arr.size[0]->kind == NAME || t->symbols[i].u.arr.size[0]->kind == NAME_FLOAT)
            fprintf(fout,"%s",t->symbols[i].u.arr.size[0]->u.name);
        else if(t->symbols[i].u.arr.size[0]->kind == CONSTANT_INT)
            fprintf(fout,"%ld",t->symbols[i].u.arr.size[0]->u.value_int);
        else if(t->symbols[i].u.arr.size[0]->kind == CONSTANT_FLOAT)
            fprintf(fout,"%0.2f",t->symbols[i].u.arr.size[0]->u.value_float);
        else{
            printf("BUG array size kind:%d\n",t->symbols[i].u.arr.size[0]->kind); 
            exit(1);
        }
        fprintf(fout,") (in the scope of the '%s' function)\n",t->symbols[i].scope_function);
      }
    }
    fprintf(fout,"#       --------\n");;
}

void symtable_free(struct symtable * t)
{
    for (unsigned int i=0 ; i<t->size; i++ ){
        if(t->symbols[i].kind == ARRAY){
            if(t->symbols[i].u.arr.size != NULL)
                free(t->symbols[i].u.arr.size);
        }
    }
    free(t->symbols);
    free(t);
}

struct code * code_new()
{
    struct code * r = malloc(sizeof(struct code));
    r->capacity = 1024;
    r->quads = malloc(r->capacity*sizeof(struct quad));
    r->nextquad = 0;
    return r;
}

static void code_grow(struct code * c)
{
    c->capacity += 1024;
    c->quads = realloc(c->quads,c->capacity*sizeof(struct quad));
    if(c->quads == NULL) 
    {
      fprintf(stderr,"Error attempting to grow quad list (actual size is %d)\n",c->nextquad);
        exit(1);
    }
}

void gencode(struct code * c,
              enum quad_kind k,
              struct symbol* s1,
              struct symbol* s2,
              struct symbol* s3)
{
    if ( c->nextquad == c->capacity )
        code_grow(c);
    struct quad * q = &(c->quads[c->nextquad]);
    q->kind = k;
    q->sym1 = s1;
    q->sym2 = s2;
    q->sym3 = s3;
    ++ (c->nextquad);
}

static void symbol_dump(struct symbol* s, FILE* fout)
{
    switch ( s->kind )
    {
        case NAME:
            fprintf(fout,"%s",s->u.name);
            break;
        case NAME_FLOAT:
            fprintf(fout,"%s",s->u.name);
            break;
        case CONSTANT_INT:
            fprintf(fout,"%ld",s->u.value_int);
            break;
        case CONSTANT_FLOAT:
            fprintf(fout,"%0.2f",s->u.value_float);
            break;
        case NAME_LOC:
            fprintf(fout,"%s",s->u.name);
            break;
        case CHAIN:
            fprintf(fout,"%s",s->u.chaine);
            break;
        case ARRAY:
            fprintf(fout,"%s",s->u.arr.name);
            break;
        case LABEL:
            fprintf(fout,"%s",s->u.name);
            break;
        default:
            fprintf(stderr,"BUG SYMBOL DUMP NOT RECOGNIZED\n");
            break;
    }
}

static void quad_dump(struct quad * q, FILE* fout)
{
    switch ( q->kind )
    {
        case BOP_PLUS:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," + ");
            symbol_dump(q->sym3,fout);
            break;
        case BOP_MINUS:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," - ");
            symbol_dump(q->sym3,fout);
            break;
        case BOP_MULT:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," * ");
            symbol_dump(q->sym3,fout);
            break;
        case BOP_DIV:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," / ");
            symbol_dump(q->sym3,fout);
            break;
        case BOP_MOD:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," %% ");
            symbol_dump(q->sym3,fout);
            break;
        case UOP_MINUS:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            fprintf(fout,"- ");
            symbol_dump(q->sym2,fout);
            break;
        case DEF_MAIN:
            symbol_dump(q->sym1,fout);
            fprintf(fout,":");
            break;
        case DEF_FUNCTION:
            symbol_dump(q->sym1,fout);
            fprintf(fout,":");
            break;
        case CALL_PRINT:
            fprintf(fout,"print ");
            symbol_dump(q->sym1,fout);
            break;
        case CALL_PRINTSTR:
            fprintf(fout,"printstr ");
            symbol_dump(q->sym1,fout);
            break;
        case CALL_READ:
            fprintf(fout,"read ");
            symbol_dump(q->sym1,fout);
            break;
        case COPY:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            break;
        case ALLOCATE_ARRAY:
            fprintf(fout,"Allocate '");
            symbol_dump(q->sym1,fout);
            fprintf(fout,"' to an one dimensional array of length ");
            symbol_dump(q->sym1->u.arr.dim1size,fout);
            break;
        case AFF_ARRAY:
            symbol_dump(q->sym1,fout);
            fprintf(fout,"[");
            symbol_dump(q->sym2,fout);
            fprintf(fout,"] := ");

            symbol_dump(q->sym3,fout);
            break;
        case COPY_ARRAY:
            symbol_dump(q->sym1,fout);
            
            fprintf(fout," := ");
            
            symbol_dump(q->sym2,fout);
            fprintf(fout,"[");
            symbol_dump(q->sym3,fout);
            fprintf(fout,"]");
            break;
        case OPREL_EQ:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," == ");
            symbol_dump(q->sym3,fout);
            break;
        case OPREL_GE:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," >= ");
            symbol_dump(q->sym3,fout);
            break;
        case OPREL_GT:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," > ");
            symbol_dump(q->sym3,fout);
            break;
        case OPREL_LE:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," <= ");
            symbol_dump(q->sym3,fout);
            break;
        case OPREL_LT:
            symbol_dump(q->sym1,fout);
            fprintf(fout," := ");
            symbol_dump(q->sym2,fout);
            fprintf(fout," < ");
            symbol_dump(q->sym3,fout);
            break;
        case LABEL_W:
            if(q->sym1!=NULL)
            {
                symbol_dump(q->sym1,fout);
                fprintf(fout," :");
            }
            break;
        case GOTO_LABEL:
            if(q->sym1!=NULL)
            {
                fprintf(fout,"goto ");
                symbol_dump(q->sym1,fout);
            }
            break;
        case ADD_INDENT:
            break;
        case RM_INDENT:
            break;
        default:
            fprintf(stderr,"BUG quad_dump(...) : q->kind inconnu\n");
            break;
    }
}

void code_dump(struct code * c, FILE* fout)
{
    unsigned int i;
    for ( i=0 ; i<c->nextquad ; i++ )
    {   
        fprintf(fout,"#%4u \t| ",i);
        quad_dump(&(c->quads[i]),fout);
        fprintf(fout,"\n");
    }
    fprintf(fout,"\n\n");
}

int indice_temp(struct symtable* t, struct symbol* s){
    unsigned int i=0;
    for(i=i; i<t->size; i++){ 
        if(s->u.name == t->symbols[i].u.name){
            return i;
        }
    }
    fprintf(stderr,"Indice_temp: s n'est pas dans la table\n");
    exit(1);

    return i;
}

void code_mips_dump(struct symtable * t, struct code * c, FILE* fout){
    fprintf(fout,".data\n");
    fprintf(fout,"\tlf: .asciiz \"\\n\"\n");
    fprintf(fout,"\tspace: .asciiz \" \"\n");
    fprintf(fout,"\ttemp_float: .float 0.0\n");
    if(t==NULL){fprintf(stderr,"Code mips dump: symtable null\n"); exit(1);}
    
    unsigned int i =0;
    for(i=i; i<t->size; i++){ 
        if(t->symbols[i].kind == CONSTANT_INT){ //INTEGER CONSTANT 
            fprintf(fout,"\ttemp%d: .word %ld\n",i,t->symbols[i].u.value_int);
        }
        else if(t->symbols[i].kind == CONSTANT_FLOAT){ //FLOAT CONSTANT
            fprintf(fout,"\ttemp%d: .float %0.2f\n",i,t->symbols[i].u.value_float);
        }
        else if(t->symbols[i].kind == CHAIN){ //CHAIN
            fprintf(fout,"\ttemp%d: .asciiz %s\n",i,t->symbols[i].u.chaine);
        }
        else if(t->symbols[i].kind == NAME_LOC){
            fprintf(fout,"\t%s: .word 0\n",t->symbols[i].u.name);
        }
        else if(t->symbols[i].kind == NAME){
            fprintf(fout,"\t%s: .word 0\n",t->symbols[i].u.name);
        }
        else if(t->symbols[i].kind == NAME_FLOAT){
            fprintf(fout,"\t%s: .float 0\n",t->symbols[i].u.name);
        } 
        else if (t->symbols[i].kind == ARRAY){
            fprintf(fout,"\t%s: .word 0\n",t->symbols[i].u.name);
        }
        // TODO determine if we can avoid using it
        // else if (t->symbols[i].kind == LABEL) {
        //     fprintf(fout,"\t%s: .word 0\n",t->symbols[i].u.name);
        // }
        else{ // TODO check whether if in else shouldn't be printed or not ?
            // fprintf(fout,"\ttemp%d: .word 0\n",i);
        }
    }
    
    #ifdef LIBDEBUG
    debug_register_strings_declaration(fout);
    #endif

    print_macros(fout);
    
    // start of the .text segment
    fprintf(fout,"\n\n.text\n");

    for(i = 0; i<c->nextquad; i++){
        switch (c->quads[i].kind)
        {
            case BOP_PLUS:
                print_bop_plus(c,t,fout,i);
                break;
            case BOP_MINUS:
                print_bop_minus(c,t,fout,i);
                break;
            case BOP_MULT:
                print_bop_mult(c,t,fout,i);
                break;
            case BOP_DIV:
                print_bop_div(c,t,fout,i);
                break;
            case BOP_MOD:
                print_bop_mod(c,t,fout,i);
                break;
            case UOP_MINUS:
                print_uop_minus(c,t,fout,i);
                break;
            case COPY_ARRAY:
                print_copy_array(t,c,fout,i);
                break;
            case ALLOCATE_ARRAY:
                print_allocate_array(t,c,fout,i);
                break;
            case AFF_ARRAY:
                print_aff_array(t,c,fout,i);
                break;
            case DEF_MAIN:
                print_main_def(fout);
                add_indent(c,t,fout);
                break;
            case CALL_PRINT:
                print_print(t,c,fout,i);
                break;
            case CALL_PRINTSTR:
                print_printstr(t,c,fout,i);
                break;
            case CALL_READ:
                print_read(t,c,fout,i);
                break;
            case COPY:
                print_copy(t,c,fout,i);
                break;
            case OPREL_EQ:
                print_EQ_check(t,c,fout,i);
                break;
            case OPREL_GT:
                print_GT_check(t,c,fout,i);
                break;
            case OPREL_GE:
                print_GE_check(t,c,fout,i);
                break;
            case OPREL_LT:
                print_LT_check(t,c,fout,i);
                break;
            case OPREL_LE:
                print_LE_check(t,c,fout,i);
                break;
            case LABEL_W:
                print_label_w(t,c,fout,i);
                break;
            case GOTO_LABEL:
                print_goto(t,c,fout,i);
                break;
            case ADD_INDENT:
                add_indent(c,t,fout);
                break;
            case RM_INDENT:
                remove_indent();
                break;
            default:
                fprintf(stderr,"#BUG %d\n",c->quads[i].kind);
                break;
        }
    }

    fprintf(fout,"\t# exit\n");
    fprintf(fout,"\tli $v0,10\n");
    fprintf(fout,"\tsyscall\n");
}

void code_free(struct code * c)
{
    free(c->quads);
    free(c);
}

void free_exit(int code, struct code* c, struct symtable* t, FILE* fname){
    if( c != NULL)
        code_free(c);
    if( t != NULL)
        symtable_free(t);
    if(fname!=NULL){fclose(fname);}
    exit(code);
}


/************************* MIPS CODE GEN FUNCTIONS *************************/

// -------- MAIN RELATED --------

void print_main_def(FILE* fout){
    fprintf(fout,".globl start\n\n");
    fprintf(fout,"start:\n");
}

// -------- ARITHMETIC RELATED --------
void print_bop_plus(struct code * c, struct symtable* t, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# start of x + y section\n",tabulation);
    #endif
    if(ISFLOAT(c->quads[i].sym2) || ISFLOAT(c->quads[i].sym3) )
    { // float minus
        load_float_bop_macro("$f0","$f1","$t0",c->quads[i].sym2,c->quads[i].sym3,t,fout);
        fprintf(fout,"%sadd.s $f2, $f0, $f1\n",tabulation);
        load_addr_macro("$t0",c->quads[i].sym1,t,fout);
        fprintf(fout,"%sswc1 $f2, 0($t0)\n",tabulation);
    } else { // integer minus
        //Check if we have variable names or just integers
        print_nameOrInteger(c,fout,i,0b110);
        //Do the operation and save result in wanted variable
        fprintf(fout,"%sadd $t0, $t0, $t1\n",tabulation);
        fprintf(fout,"%ssw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
    }
    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of x + y section\n",tabulation);
    #endif
}


void print_bop_minus(struct code * c, struct symtable* t, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# start of x - y section\n",tabulation);
    #endif
    if(ISFLOAT(c->quads[i].sym2) || ISFLOAT(c->quads[i].sym3) )
    { // float minus
        load_float_bop_macro("$f0","$f1","$t0",c->quads[i].sym2,c->quads[i].sym3,t,fout);
        fprintf(fout,"%ssub.s $f2, $f0, $f1\n",tabulation);
        load_addr_macro("$t0",c->quads[i].sym1,t,fout);
        fprintf(fout,"%sswc1 $f2, 0($t0)\n",tabulation);
    } else { // integer minus
        //Check if we have variable names or just integers
        print_nameOrInteger(c,fout,i,0b110);
        //Do the operation and save result in wanted variable
        fprintf(fout,"%ssub $t0, $t0, $t1\n",tabulation);
        fprintf(fout,"%ssw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
    }
    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of x - y section\n",tabulation);
    #endif
}

void print_bop_mult(struct code * c, struct symtable* t, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# start of x * y section\n",tabulation);
    #endif
    if(ISFLOAT(c->quads[i].sym2) || ISFLOAT(c->quads[i].sym3) )
    { // float multiplication
        load_float_bop_macro("$f0","$f1","$t0",c->quads[i].sym2,c->quads[i].sym3,t,fout);
        fprintf(fout,"%smul.s $f2, $f0, $f1\n",tabulation);
        load_addr_macro("$t0",c->quads[i].sym1,t,fout);
        fprintf(fout,"%sswc1 $f2, 0($t0)\n",tabulation);
    } else { // integer multiplication
        //Check if we have variable names or just integers for sym2 and sym3
        print_nameOrInteger(c,fout,i,0b110);
        //Do the operation and save result in wanted variable
        fprintf(fout,"%smul $t0, $t0, $t1\n",tabulation);
        fprintf(fout,"%ssw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
    }
    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of x * y section\n",tabulation);
    #endif
}

void print_bop_div(struct code * c, struct symtable* t, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# start of x / y section\n",tabulation);
    #endif
    if(ISFLOAT(c->quads[i].sym2) || ISFLOAT(c->quads[i].sym3) )
    { // float division
        load_float_bop_macro("$f0","$f1","$t0",c->quads[i].sym2,c->quads[i].sym3,t,fout);
        fprintf(fout,"%sdiv.s $f2, $f0, $f1\n",tabulation);
        load_addr_macro("$t0",c->quads[i].sym1,t,fout);
        fprintf(fout,"%sswc1 $f2, 0($t0)\n",tabulation);
    } else { // integer division
        //Check if we have variable names or just integers for sym2 and sym3
        print_nameOrInteger(c,fout,i,0b110);
        //Do the operation and save result in wanted variable
        fprintf(fout,"%sdiv $t0, $t0, $t1\n",tabulation);
        fprintf(fout,"%ssw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
    }
    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of x / y section\n",tabulation);
    #endif
}

void print_bop_mod(struct code * c, struct symtable* t, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# start of x %% y section\n",tabulation);
    #endif
    if(ISFLOAT(c->quads[i].sym2) || ISFLOAT(c->quads[i].sym3) )
    { // float modulo
        load_float_bop_macro("$f0","$f1","$t0",c->quads[i].sym2,c->quads[i].sym3,t,fout);
        // get the quotient 
        fprintf(fout,"%sdiv.s $f2, $f0, $f1\n",tabulation);
        // floor it down
        fprintf(fout,"%sfloor.w.s $f3, $f2 \n",tabulation);
        // the above operation turned it into the integer format we need to turn it back to a float
        fprintf(fout,"%scvt.s.w $f3, $f3 \n",tabulation);
        // we now multiply divisor and the obtained quotient
        fprintf(fout,"%smul.s $f4, $f3, $f1\n",tabulation);
        // we substract this to the dividend we obtain the remainder of the division hence we have the mod
        fprintf(fout,"%ssub.s $f5, $f0, $f4\n",tabulation);
        load_addr_macro("$t0",c->quads[i].sym1,t,fout);
        fprintf(fout,"%sswc1 $f5, 0($t0)\n",tabulation);
    } else { // integer modulo
        //Check if we have variable names or just integers for sym2 and sym3
        print_nameOrInteger(c,fout,i,0b110);
        //Do the operation and save result in wanted variable
        fprintf(fout,"%sdiv $t0, $t0, $t1\n",tabulation);
        fprintf(fout,"%smfhi $t0\n",tabulation);
        fprintf(fout,"%ssw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
    }
    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of x %% y section\n",tabulation);
    #endif
}

void print_uop_minus(struct code * c, struct symtable* t, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# start of -x section\n",tabulation);
    #endif
    if(ISFLOAT(c->quads[i].sym2))
    { // float uop minus
        if(!ISFLOAT(c->quads[i].sym2))
        { // if sym2 is an integer convert it to a float and store it in $f0
                convert_int_to_float("$f0","$t0",c->quads[i].sym2,t,fout);
        } else { // store it in $f0
                load_symbol_macro("$f0","$t0",c->quads[i].sym2,t,fout);
        }
        fprintf(fout,"%sneg.s $f1, $f0\n",tabulation);
        load_addr_macro("$t0",c->quads[i].sym1,t,fout);
        fprintf(fout,"%sswc1 $f1, 0($t0)\n",tabulation);
    } else {
        //Check if we have variable names or just integers
        if(c->quads[i].sym2->kind == NAME || c->quads[i].sym2->kind == NAME_LOC)
            fprintf(fout,"%slw $t0, %s\n",tabulation,c->quads[i].sym2->u.name);
        else if(c->quads[i].sym2->kind == CONSTANT_INT)
            fprintf(fout,"%sli $t0, %ld\n",tabulation,c->quads[i].sym2->u.value_int);

        //Do the operation and save result in wanted variable
        fprintf(fout,"%sneg $t1, $t0\n",tabulation);
        fprintf(fout,"%ssw $t1, %s\n",tabulation,c->quads[i].sym1->u.name);
    }
    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of -x section\n",tabulation);
    #endif
}

// -------- COMPARISONS / OPREL RELATED --------

void print_EQ_check(struct symtable* t, struct code * c, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# equality check\n",tabulation);
    #endif
    if(c->quads[i].sym1->kind!=LABEL)
    {
        fprintf(stdout,"error gave a non label symbol, exiting\n");
        free_exit(1,c,t,fout);
    }
    if(ISFLOAT(c->quads[i].sym2)|| ISFLOAT(c->quads[i].sym3))
    {
        load_float_bop_macro("$f0","$f1","$t0",c->quads[i].sym2,c->quads[i].sym3,t,fout);
        fprintf(fout,"%sc.eq.s $f0, $f1\n",tabulation);
        fprintf(fout,"%sbc1f %s\n",tabulation,c->quads[i].sym1->u.name);
    } else {
        // Check if we have variable names or just integers for sym2 and sym3
        print_nameOrInteger(c,fout,i,0b110);
        // we wanna branch to the else when condition is false s owe do the opposite test == => !=
        fprintf(fout,"%sbne $t0, $t1, %s\n",tabulation,c->quads[i].sym1->u.name);
    }
}

void print_GT_check(struct symtable* t, struct code * c, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# Greater than check\n",tabulation);
    #endif
    if(c->quads[i].sym1->kind!=LABEL)
    {
        fprintf(stdout,"error gave a non label symbol, exiting\n");
        free_exit(1,c,t,fout);
    }
    if(ISFLOAT(c->quads[i].sym2)|| ISFLOAT(c->quads[i].sym3))
    {
        load_float_bop_macro("$f0","$f1","$t0",c->quads[i].sym2,c->quads[i].sym3,t,fout);
        // we go to else if lesser or equal is true since it's the oposite of greater than
        fprintf(fout,"%sc.le.s $f0, $f1\n",tabulation);
        fprintf(fout,"%sbc1t %s\n",tabulation,c->quads[i].sym1->u.name);
    } else {
        // Check if we have variable names or just integers for sym2 and sym3
        print_nameOrInteger(c,fout,i,0b110);
        // we go to else if lesser or equal is true since it's the oposite of greater than
        fprintf(fout,"%sble $t0, $t1, %s\n",tabulation,c->quads[i].sym1->u.name);
    }
}
void print_GE_check(struct symtable* t, struct code * c, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# Greater or equal check\n",tabulation);
    #endif
    if(c->quads[i].sym1->kind!=LABEL)
    {
        fprintf(stdout,"error gave a non label symbol, exiting\n");
        free_exit(1,c,t,fout);
    }
    if(ISFLOAT(c->quads[i].sym2)|| ISFLOAT(c->quads[i].sym3))
    {
        load_float_bop_macro("$f0","$f1","$t0",c->quads[i].sym2,c->quads[i].sym3,t,fout);
        // we go to else if lesser than is true since it's the oposite of greater or equal
        fprintf(fout,"%sc.lt.s $f0, $f1\n",tabulation);
        fprintf(fout,"%sbc1t %s\n",tabulation,c->quads[i].sym1->u.name);
    } else {
        // Check if we have variable names or just integers for sym2 and sym3
        print_nameOrInteger(c,fout,i,0b110);
        // we go to else if lesser than is true since it's the oposite of greater or equal
        fprintf(fout,"%sblt $t0, $t1, %s\n",tabulation,c->quads[i].sym1->u.name);
    }
}
void print_LT_check(struct symtable* t, struct code * c, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# Lesser than check\n",tabulation);
    #endif
    if(c->quads[i].sym1->kind!=LABEL)
    {
        fprintf(stdout,"error gave a non label symbol, exiting\n");
        free_exit(1,c,t,fout);
    }
    if(ISFLOAT(c->quads[i].sym2)|| ISFLOAT(c->quads[i].sym3))
    {
        load_float_bop_macro("$f0","$f1","$t0",c->quads[i].sym2,c->quads[i].sym3,t,fout);
        // we go to else if lesser than is false
        fprintf(fout,"%sc.lt.s $f0, $f1\n",tabulation);
        fprintf(fout,"%sbc1f %s\n",tabulation,c->quads[i].sym1->u.name);
    } else {
        // Check if we have variable names or just integers for sym2 and sym3
        print_nameOrInteger(c,fout,i,0b110);
        // we go to else if greater or equal is true since it's the oposite of lesser than
        fprintf(fout,"%sbge $t0, $t1, %s\n",tabulation,c->quads[i].sym1->u.name);
    }
}
void print_LE_check(struct symtable* t, struct code * c, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# Lesser or equal check\n",tabulation);
    #endif
    if(c->quads[i].sym1->kind!=LABEL)
    {
        fprintf(stdout,"error gave a non label symbol, exiting\n");
        free_exit(1,c,t,fout);
    }
    if(ISFLOAT(c->quads[i].sym2)|| ISFLOAT(c->quads[i].sym3))
    {
        load_float_bop_macro("$f0","$f1","$t0",c->quads[i].sym2,c->quads[i].sym3,t,fout);
        // we go to else if lesser or equal is false
        fprintf(fout,"%sc.le.s $f0, $f1\n",tabulation);
        fprintf(fout,"%sbc1f %s\n",tabulation,c->quads[i].sym1->u.name);
    } else {
        // Check if we have variable names or just integers for sym2 and sym3
        print_nameOrInteger(c,fout,i,0b110);
        // we go to else if greater than is true since it's the oposite of lesser or equal
        fprintf(fout,"%sbgt $t0, $t1, %s\n",tabulation,c->quads[i].sym1->u.name);
    }
}

// -------- BRANCHING RELATED --------

void print_label_w(struct symtable* t, struct code * c, FILE* fout,int i)
{
    if(c->quads[i].sym1->kind!=LABEL)
    {
        fprintf(stdout,"error gave a non label symbol, exiting\n");
        free_exit(1,c,t,fout);
    }
    fprintf(fout,"%s%s:\n",tabulation,c->quads[i].sym1->u.name);
}

void print_goto(struct symtable* t, struct code * c, FILE* fout,int i)
{
    if(c->quads[i].sym1->kind!=LABEL)
    {
        fprintf(stdout,"error gave a non label symbol, exiting\n");
        free_exit(1,c,t,fout);
    }
    fprintf(fout,"%sj %s\n",tabulation,c->quads[i].sym1->u.name);
}
// -------- PRINT RELATED --------

void print_print(struct symtable * t,struct code * c, FILE* fout,int i)
{
    switch(c->quads[i].sym1->kind){
        case ARRAY:
            // TODO + WIP solution temporaire, quand les fonctions seront implémentés à remplacer par une fonction pour éviter le code spaghetti
            //  putting the array size in $t2
            load_symbol_macro("$t2","NULL",c->quads[i].sym1->u.arr.dim1size,t,fout);
            // to avoid doing a mult at each iterations we multiply the size by 4 (each elements need a 4 bytes incrementation)
            fprintf(fout,"%smul $t0, $t2, 4\n",tabulation);
            //  putting array address in $t1
            load_symbol_macro("$t1","NULL",c->quads[i].sym1,t,fout);
            fprintf(fout,"%sli $t2, 0\n",tabulation);


            // we avoid having 2 identical labels by putting the quad index at the end of it
            // /!\ start of the loop /!\ --
            fprintf(fout,"%sprint_array_loop%d:\n",tabulation,i);
            add_indent(c,t,fout);
            // bge by security to avoid miss matching (which shouldn't be happening)
            fprintf(fout,"%sbge $t2, $t0, end_array_loop%d\n",tabulation,i);
            
            // put the address of the current element to be printed in $t3
            fprintf(fout,"%sadd $t3, $t2, $t1\n",tabulation);
            // load the current float element in the
            fprintf(fout,"%slwc1 $f0, ($t3)\n",tabulation);
            // function that print the asked register
            print_reg("f0",fout,FLOAT);
            // macro that print a space
            fprintf(fout,"%sSPACE\n",tabulation);
            // increment to next address
            fprintf(fout,"%saddi $t2, $t2, 4\n",tabulation);

            // jump to the start of the loop
            fprintf(fout,"%sj print_array_loop%d\n",tabulation,i);
            // /!\ end of the loop /!\ --
            remove_indent();
            fprintf(fout,"%send_array_loop%d:\n",tabulation,i);
            fprintf(fout,"%sLF\n",tabulation);
            break;
        default:
            fprintf(fout,"%slw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
            fprintf(fout,"%smove $a0, $t0\n",tabulation);
            // ne marche que pour print ls int pas pour les float
            fprintf(fout,"%sli $v0, 1\n",tabulation);
            fprintf(fout,"%ssyscall\n",tabulation);
            //Return to line
            fprintf(fout,"%sLF\n",tabulation);
            break;
    }
    
}

void print_printstr(struct symtable * t, struct code * c, FILE* fout,int i)
{
    fprintf(fout,"%sli $v0, 4\n",tabulation);
    fprintf(fout,"%sla $a0, temp%d\n",tabulation,indice_temp(t,c->quads[i].sym1));
    fprintf(fout,"%ssyscall\n",tabulation);
}


void print_reg(char * reg,FILE * fout, enum var_type ty){
    switch(ty) {
        case INTEGER:
            fprintf(fout,"%smove $a0, $%s\n",tabulation,reg);
            fprintf(fout,"%sli $v0, 1\n",tabulation);
            break;
        case FLOAT:
            fprintf(fout,"%smov.s $f12, $%s\n",tabulation,reg);
            fprintf(fout,"%sli $v0, 2\n",tabulation);
            break;
        default:
            // we suppose value is int
            fprintf(fout,"%smove $a0, $%s\n",tabulation,reg);
            fprintf(fout,"%sli $v0, 1\n",tabulation);
            break;
    }
    fprintf(fout,"%ssyscall\n",tabulation);
}

// -------- READ RELATED --------

void print_read(struct symtable * t,struct code * c, FILE* fout,int i)
{
    switch(c->quads[i].sym1->kind){
        case ARRAY:
            #ifdef LIBDEBUG
            fprintf(fout,"%s# float read section\n",tabulation);
            #endif
            // System call read float
            fprintf(fout,"%sli $v0, 6\n",tabulation);
            fprintf(fout,"%ssyscall\n",tabulation);
            // Assign given value to float variable
            fprintf(fout,"%slw $t2, %s\n",tabulation,c->quads[i].sym1->u.name);
            fprintf(fout,"%saddi $t0, $t2, 0\n",tabulation);
            fprintf(fout,"%sswc1 $f0, ($t0)\n",tabulation);
            #ifdef LIBDEBUG
            fprintf(fout,"%s# end of float read section\n",tabulation);
            #endif
            break;
        case NAME_LOC:
            // System call read int
            fprintf(fout,"%sli $v0, 5\n",tabulation);
            fprintf(fout,"%ssyscall\n",tabulation);
            // Assign given value to int variable
            fprintf(fout,"%ssw $v0, %s\n",tabulation,c->quads[i].sym1->u.name);
            break;
        default:
            fprintf(stderr,"Can not assign value into this kind\n");
            break;
    }

}

// -------- FUNCTIONS RELATED --------


// -------- ARRAY RELATED --------

void print_copy_array(struct symtable * t, struct code * c, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# copy_array section\n",tabulation);
    #endif
    if(c->quads[i].sym2->kind != ARRAY){
        fprintf(stdout,"Tried to do an  Var = tab[index] operation but right part isn't an array should address this issue in the corec.y file\n");
        free_exit(1,c,t,fout);
    }
    // adress of the array element we want to copy
    load_tab_addr_macro("$t0", "$t1", "$t2", c->quads[i].sym2,c->quads[i].sym3,t,fout);
    // adress of the place we want to copy the array element in
    load_addr_macro("$t1",c->quads[i].sym1,t,fout);
    // load in $f0 what's inside the array element
    fprintf(fout,"%slwc1 $f0, ($t0)\n",tabulation);
    // store it in the destination
    fprintf(fout,"%sswc1 $f0, ($t1)\n",tabulation);

    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of copy_array section\n",tabulation);
    #endif
}

void print_allocate_array(struct symtable * t, struct code * c, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# allocate_array section\n",tabulation);
    #endif
    if(c->quads[i].sym1->kind != ARRAY){
        fprintf(stdout,"Tried to don an tab[index] = Var operation but left part isn't an array should address this issue in the corec.y file\n");
        free_exit(1,c,t,fout);
    }
    // number for the sbrk primitive call
    fprintf(fout,"%sli $v0, 9\n",tabulation);
    // based on the type of the var indicating the size of the array we do different operations to end up with the nb of bytes we should allocate is $a0
    load_symbol_alloc_macro("$a0","$t0",c->quads[i].sym1->u.arr.dim1size,t,fout);
    // call sbrk
    fprintf(fout,"%ssyscall\n",tabulation);
    // save the allocated array ptr returned in $v0 by syscall in the assigned .data var
    fprintf(fout,"%sla $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
    fprintf(fout,"%ssw $v0, ($t0)\n",tabulation);
    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of allocate_array section\n",tabulation);
    #endif
}

void print_aff_array(struct symtable * t, struct code * c, FILE* fout,int i)
{
    #ifdef LIBDEBUG
    fprintf(fout,"%s# aff_array section\n",tabulation);
    #endif
    if(c->quads[i].sym1->kind != ARRAY){
        fprintf(stdout,"Tried to do an tab[index] = Var operation but left part isn't an array should address this issue in the corec.y file\n");
        free_exit(1,c,t,fout);
    }
    // V alternative, effectue directement le*4 en calcul vu qu'on devrait passer par un load immediate dans tout les cas 
    // fprintf(fout,"%sla $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
    load_tab_addr_macro("$t0", "$t1", "$t2", c->quads[i].sym1,c->quads[i].sym2,t,fout);
    
    if(!ISFLOAT(c->quads[i].sym3))
    {   // TODO make this temporary fix cleaner
        convert_int_to_float("$f0","$t2",c->quads[i].sym3,t,fout);
        fprintf(fout,"%smfc1 $t2, $f0\n",tabulation);
    } else
        load_symbol_macro("$f0","$t1",c->quads[i].sym3,t,fout); // sym3 is a float so the tmp register will be
    fprintf(fout,"%sswc1 $f0, ($t0)\n",tabulation);
    #ifdef LIBDEBUG
    fprintf(fout,"%s# end of aff_array section\n",tabulation);
    #endif
}   




// -------- MISCELLANEOUS --------

void print_copy(struct symtable * t, struct code * c, FILE* fout,int i)
{
    if(c->quads[i].sym2->kind == NAME || c->quads[i].sym2->kind == NAME_LOC)
        fprintf(fout,"%slw $t0, %s\n",tabulation,c->quads[i].sym2->u.name);
    else
        fprintf(fout,"%slw $t0, temp%d\n",tabulation,indice_temp(t,c->quads[i].sym2));
    fprintf(fout,"%ssw $t0, %s\n",tabulation,c->quads[i].sym1->u.name);
}

/***    do the right thing based on whether we have an named var or a constant TODO finde a better name !
 *      if on sym 1 sym = 0b1 do on sym1 (not implmented should never be of use)
 *      if on sym 2 sym = 0b10 etc
 *      if on sym 2 and 3 but not 1 simultaneously sym = 0b110
 *      to use only on BOP operations PLUS MINUS MUL DIV MOD
 *      TODO sym may be useless and may be removed later
 */ 

void print_nameOrInteger(struct code * c, FILE* fout,int i,int sym)
{
    if(sym & 0b10) {
        if(c->quads[i].sym2->kind == NAME || c->quads[i].sym2->kind == NAME_LOC)
            fprintf(fout,"%slw $t0, %s\n",tabulation,c->quads[i].sym2->u.name);
        else if(c->quads[i].sym2->kind == CONSTANT_INT)
            fprintf(fout,"%sli $t0, %ld\n",tabulation,c->quads[i].sym2->u.value_int);
    }
    if( sym & 0b100) {
        if(c->quads[i].sym3->kind == NAME || c->quads[i].sym3->kind == NAME_LOC)
            fprintf(fout,"%slw $t1, %s\n",tabulation,c->quads[i].sym3->u.name);
        else if(c->quads[i].sym3->kind == CONSTANT_INT)
            fprintf(fout,"%sli $t1, %ld\n",tabulation,c->quads[i].sym3->u.value_int);
    }
}


// -------- MACROS --------

void load_symbol_alloc_macro(char * dest, char * tmp_used, struct symbol * sym, struct symtable * t, FILE* fout){
    if(sym->kind == CONSTANT_INT) // var = 3
        // in this case the value is a constant int we can do a load immediate in $a0 with value*4
        fprintf(fout,"%sli %s, %ld\n",tabulation,dest,(sym->u.value_int)*4);
    // in other cases we need to load the value from the label of the corresponding symbol
    else if( sym->kind == NAME || sym->kind == NAME_LOC ) { // var = 4*3
        fprintf(fout,"%slw %s, %s\n",tabulation,tmp_used,(sym->u.name));
    } else { // array[4] = array[5]+3*4
        fprintf(fout,"%slw %s, temp%d\n",tabulation,tmp_used,indice_temp(t,sym));
    }
    // if we don't have a CONST INT we loaded the nb of int/floats we want to stock in the array, in this case
    if(sym->kind  != CONSTANT_INT)
        fprintf(fout,"%ssll %s, %s, 2\n",tabulation,dest,tmp_used);
}

void load_symbol_macro(char * dest, char * tmp_used, struct symbol * sym, struct symtable * t, FILE* fout){
    if(sym->kind == CONSTANT_INT) // var = 3
        // in this case the value is a constant int we can do a load immediate in $a0 with value*4
        fprintf(fout,"%sli %s, %ld\n",tabulation,dest,(sym->u.value_int));
    // in other cases we need to load the value from the label of the corresponding symbol
    else if( sym->kind == NAME || sym->kind == NAME_LOC || sym->kind == ARRAY) { // var = 4*3
        fprintf(fout,"%slw %s, %s\n",tabulation,dest,(sym->u.name));
    } else if (sym->kind == NAME_FLOAT || sym->kind == CONSTANT_FLOAT) {
        if( sym->kind == NAME_FLOAT )
            fprintf(fout,"%sla %s, %s\n",tabulation,tmp_used,(sym->u.name));
        else
            fprintf(fout,"%sla %s, temp%d\n",tabulation,tmp_used,indice_temp(t,sym));
        fprintf(fout,"%slwc1 %s, 0(%s)\n",tabulation,dest,tmp_used);
    } else { // array[4] = array[5]+3*4
        fprintf(fout,"%slw %s, temp%d\n",tabulation,dest,indice_temp(t,sym));
    }
}


void load_tab_addr_macro(char * dest,char * tmp_used1, char * tmp_used2, struct symbol * sym_addr, struct symbol * sym_index, struct symtable * t, FILE* fout){
    fprintf(fout,"%slw %s, %s\n",tabulation,tmp_used2,sym_addr->u.name);
    if(sym_index->kind == CONSTANT_INT)
        // put into dest the addr of array[index] element
        fprintf(fout,"%saddi %s, %s, %ld\n",tabulation,dest,tmp_used2,(sym_index->u.value_int)*4);
    // in other cases we need to load the value from the label of the corresponding symbol
    else if( sym_index->kind == NAME || sym_index->kind == NAME_LOC || sym_index->kind == ARRAY) {
        fprintf(fout,"%slw %s, %s\n",tabulation,tmp_used1,(sym_index->u.name));
    } else {
        fprintf(fout,"%slw %s, temp%d\n",tabulation,tmp_used1,indice_temp(t,sym_index));
    }
    if (sym_index->kind != CONSTANT_INT){
        // if not CONSTANT_INT wasn't yet multiplied by 4
        fprintf(fout,"%ssll %s, %s, 2\n",tabulation, tmp_used1, tmp_used1);
        fprintf(fout,"%sadd %s, %s, %s\n",tabulation, dest, tmp_used2, tmp_used1);
    }
}

void load_addr_macro(char * dest, struct symbol * sym, struct symtable * t, FILE* fout)
{
    if( sym->kind == NAME || sym->kind == NAME_FLOAT || sym->kind == NAME_LOC || sym->kind == ARRAY){
        fprintf(fout,"%sla %s, %s\n",tabulation,dest,sym->u.name);
    } else 
        fprintf(fout,"%sla %s, temp%d\n",tabulation,dest,indice_temp(t,sym));

}

void convert_int_to_float(char * dest,char * tmp, struct symbol * sym, struct symtable * t, FILE* fout)
{
    load_addr_macro(tmp,sym,t,fout);
    fprintf(fout,"%slw %s, 0(%s)\n",tabulation,tmp,tmp);
    fprintf(fout,"%smtc1 %s, %s\n",tabulation,tmp,dest);
    // we convert the integer to a float
    fprintf(fout,"%scvt.s.w %s, %s\n",tabulation,dest,dest); // the value in dest now is a float
}



void load_float_bop_macro(char * dest1, char* dest2, char * tmp, struct symbol * sym1, struct symbol * sym2, struct symtable * t, FILE* fout){
    if(!ISFLOAT(sym1))
    { // if sym2 is an integer convert it to a float and store it in $f0
            convert_int_to_float(dest1,tmp,sym1,t,fout);
    } else { // store it in $f0
            load_symbol_macro(dest1,tmp,sym1,t,fout);
    }
    if(!ISFLOAT(sym2))
    { // if sym2 is an integer convert it to a float and store it in $f1
            convert_int_to_float(dest2,tmp,sym2,t,fout);
    } else { // store it in $f1
            load_symbol_macro(dest2,tmp,sym2,t,fout);
    }
}

void print_macros(FILE* fout){
    fprintf(fout,"\n# MACROS definition\n\n");
    print_LF_macro(fout);
    print_SPACE_macro(fout);
    fprintf(fout,"# End of MACROS definition\n");
}

void print_LF_macro(FILE* fout){
    fprintf(fout,".macro LF\n");
    fprintf(fout,"\tli $v0, 4\n");
    fprintf(fout,"\tla $a0, lf\n");
    fprintf(fout,"\tsyscall\n");
    fprintf(fout,".end_macro\n");
}

void print_SPACE_macro(FILE* fout){
    fprintf(fout,".macro SPACE\n");
    fprintf(fout,"\tli $v0, 4\n");
    fprintf(fout,"\tla $a0, space\n");
    fprintf(fout,"\tsyscall\n");
    fprintf(fout,".end_macro\n");
}

// ETC FEEL FREE TO ADD MORE SECTIONS

// -------- DEBUG RELATED --------

#ifdef LIBDEBUG
void debug_print_reg(char * reg,FILE * fout, int type){
    fprintf(fout,"%sli $v0, 4\n",tabulation);
    fprintf(fout,"%sla $a0, print%s\n",tabulation,reg);
    fprintf(fout,"%ssyscall\n",tabulation);
    
    fprintf(fout,"%smove $a0, $%s\n",tabulation,reg);
    switch(type) {
        case 0:
            fprintf(fout,"%sli $v0, 1\n",tabulation);
            break;
        case 1:
            fprintf(fout,"%sli $v0, 4\n",tabulation);
            break;
        default:
            // we suppose value is int
            fprintf(fout,"%sli $v0, 1\n",tabulation);
            break;
    }
    fprintf(fout,"%ssyscall\n",tabulation);
    //Return to line
    fprintf(fout,"%sLF\n",tabulation);

}

void debug_print_start(FILE * fout){
    fprintf(fout,"%sli $v0, 4\n",tabulation);
    fprintf(fout,"%sla $a0, startfct\n",tabulation);
    fprintf(fout,"%ssyscall\n",tabulation);
    fprintf(fout,"%sLF\n",tabulation);
}

void debug_print_end(FILE * fout){
    fprintf(fout,"%sli $v0, 4\n",tabulation);
    fprintf(fout,"%sla $a0, endfct\n",tabulation);
    fprintf(fout,"%ssyscall\n",tabulation);
    fprintf(fout,"%sLF\n",tabulation);
}

void debug_register_strings_declaration(FILE * fout){
    for(int i = 0; i < 10; i++)
        fprintf(fout,"\tprintt%d: .asciiz \"$t%d : \"\n",i,i);
    for(int i = 0; i < 4; i++)
        fprintf(fout,"\tprinta%d: .asciiz \"$a%d : \"\n",i,i);
    for(int i = 0; i < 3; i++)
        fprintf(fout,"\tprintv%d: .asciiz \"$v%d : \"\n",i,i);
    fprintf(fout,"\tstartfct: .asciiz \"START \"\n" );
    fprintf(fout,"\tendfct: .asciiz \"END \"\n" );
    
}
#endif
// -------- ERROR CHECK --------

void push_stack(labelStack ** stack, wardLabels new_elt)
{
    labelStack * new;
    CHK( new = malloc(sizeof(labelStack)) , NULL );
    new->elt.cond_number = new_elt.cond_number;
    new->elt.lbl1 = new_elt.lbl1;
    new->elt.lbl2 = new_elt.lbl2;
    new->previous = (*stack);
    (*stack) = new;
}

void pop_stack(labelStack ** stack)
{
    if((*stack) == NULL)
    {
        fprintf(stdout,"Trying to pop an empty stack \n");
        return;
    }
    labelStack * tmp = (*stack);
    (*stack) = tmp->previous;
    free(tmp);
}

void raler(int err)
{
    if (err == 1)
        perror ("");
    exit(EXIT_FAILURE);

}

/************************* NAME TO DEFINE FUNCTIONS *************************/



void add_indent(struct code* c, struct symtable* t, FILE* fname){
    if(indent >= 1022)
    {
        printf("more than 1023 indentations ! write better code.\n");
        free_exit(1,c,t,fname);
    }
    tabulation[indent] = '\t';
    indent++;
    tabulation[indent] = '\0';
}

int remove_indent(){
    if(indent == 0)
    {
        printf("shouldn't be removing a tab since there are none, resuming \n");
        return -1;
    }
    indent--;
    tabulation[indent] = '\0';
    return 0;
}

void init_indent()
{
    indent=0;
    tabulation[indent]='\0';
}